/***********************************************************************************************************************
 *
 * blueBill Core - open source birding
 * Copyright (C) 2009-2011 by Tidalwave s.a.s. (http://www.tidalwave.it)
 *
 ***********************************************************************************************************************
 *
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on
 * an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  See the License for the
 * specific language governing permissions and limitations under the License.
 *
 ***********************************************************************************************************************
 *
 * WWW: http://bluebill.tidalwave.it
 * SCM: https://java.net/hg/bluebill~core-src
 *
 **********************************************************************************************************************/
package it.tidalwave.semantic.io.impl;

import it.tidalwave.semantic.io.GraphSerializer;
import it.tidalwave.semantic.io.json.JsonPrettyPrinterStream;
import it.tidalwave.semantic.io.json.RdfJsonWriter;
import javax.annotation.Nonnull;
import java.io.IOException;
import java.io.OutputStream;
import java.io.OutputStreamWriter;
import java.io.Writer;
import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;
import java.util.TreeMap;
import org.openrdf.model.Graph;
import org.openrdf.model.Statement;
import org.openrdf.rio.RDFHandlerException;
import org.openrdf.rio.RDFWriter;
import org.openrdf.rio.n3.N3Writer;
import org.openrdf.rio.rdfxml.util.RDFXMLPrettyWriter;
import static it.tidalwave.semantic.io.MimeTypes.*;

/***********************************************************************************************************************
 *
 * @author  Fabrizio Giudici
 * @version $Id$
 *
 **********************************************************************************************************************/
public class DefaultGraphSerializer implements GraphSerializer
  {
    private final Map<String, String> uriMapByPrefix = new TreeMap<String, String>();
    private final StatementComparator statementComparator = new StatementComparator();

    /*******************************************************************************************************************
     *
     * {@inheritDoc}
     *
     ******************************************************************************************************************/
    public void write (final @Nonnull Graph graph, @Nonnull OutputStream os, final @Nonnull String mimeType)
      throws IOException, RDFHandlerException
      {
        if (MIME_RDF_JSON.equals(mimeType))
          {
            os = new JsonPrettyPrinterStream(os);
          }

        write(graph, new OutputStreamWriter(os, "UTF-8"), mimeType);
      }

    /*******************************************************************************************************************
     *
     * {@inheritDoc}
     *
     ******************************************************************************************************************/
    public void write (final @Nonnull Graph graph, final @Nonnull Writer w, final @Nonnull String mimeType)
      throws IOException, RDFHandlerException
      {
        RDFWriter rdfWriter = null;

        // We're not using yet the registry since we don't know whether it works with Android
        if (MIME_RDF_JSON.equals(mimeType))
          {
            rdfWriter = new RdfJsonWriter(w);
          }
        else if (MIME_RDF_XML.equals(mimeType))
          {
            rdfWriter = new RDFXMLPrettyWriter(w);
          }
        else if (MIME_N3.equals(mimeType))
          {
            rdfWriter = new N3Writer(w);
          }
        else
          {
            throw new RDFHandlerException("No writer for MIME type: " + mimeType);
          }

        for (final Entry<String, String> e : uriMapByPrefix.entrySet())
          {
            rdfWriter.handleNamespace(e.getKey(), e.getValue());
          }

        write(graph, rdfWriter);
        w.flush(); // some writers (e.g. json) seems to need it
      }

    /*******************************************************************************************************************
     *
     * {@inheritDoc}
     *
     ******************************************************************************************************************/
    public void handleNamespace (final @Nonnull String prefix, final @Nonnull String uri)
      {
        uriMapByPrefix.put(prefix, uri);
      }

    /*******************************************************************************************************************
     *
     * {@inheritDoc}
     *
     ******************************************************************************************************************/
    private void write (final @Nonnull Graph graph, final @Nonnull RDFWriter writer)
      throws RDFHandlerException
      {
        writer.startRDF();

        for (final Statement statement : sortedAndUnique(graph))
          {
            writer.handleStatement(statement);
          }

        writer.endRDF();
      }

    /*******************************************************************************************************************
     *
     *
     ******************************************************************************************************************/
    @Nonnull
    private List<Statement> sortedAndUnique (final @Nonnull Collection<Statement> statements)
      {
        final List<Statement> result = new ArrayList<Statement>(statements);

        if (!result.isEmpty())
          {
            Collections.sort(result, statementComparator);
            final Iterator<Statement> i = result.iterator();
            Statement previous = i.next();

            while (i.hasNext())
              {
                final Statement statement = i.next();

                if (statementComparator.compare(statement, previous) == 0)
                  {
                    i.remove();
                  }

                previous = statement;
              }
          }

        return result;
      }
  }
