/***********************************************************************************************************************
 *
 * blueBill Core - open source birding
 * Copyright (C) 2009-2011 by Tidalwave s.a.s. (http://www.tidalwave.it)
 *
 ***********************************************************************************************************************
 *
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on
 * an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  See the License for the
 * specific language governing permissions and limitations under the License.
 *
 ***********************************************************************************************************************
 *
 * WWW: http://bluebill.tidalwave.it
 * SCM: https://java.net/hg/bluebill~core-src
 *
 **********************************************************************************************************************/
package it.tidalwave.semantic.io.impl;

import javax.annotation.Nonnull;
import javax.annotation.concurrent.NotThreadSafe;
import java.util.ArrayList;
import java.util.Collection;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.SortedMap;
import java.util.Stack;
import java.util.TreeMap;
import it.tidalwave.util.As;
import it.tidalwave.util.Id;
import it.tidalwave.util.NotFoundException;
import it.tidalwave.semantic.Type;
import it.tidalwave.semantic.io.GraphUnmarshaller;
import it.tidalwave.semantic.io.spi.StatementUnmarshaller;
import it.tidalwave.semantic.io.spi.StatementUnmarshallerFactory;
import it.tidalwave.util.StringValue;
import org.openide.util.Lookup;
import org.openrdf.model.Statement;
import org.openrdf.model.Graph;
import org.openrdf.model.Literal;
import org.openrdf.model.URI;

/***********************************************************************************************************************
 *
 * @author  Fabrizio Giudici
 * @version $Id$
 *
 **********************************************************************************************************************/
@NotThreadSafe
public class GraphUnmarshallerImpl implements GraphUnmarshaller, StatementUnmarshaller.Context
  {
    private Graph graph;

    private final Map<Id, Object> entities = new HashMap<Id, Object>();

    private SortedMap<Id, List<Statement>> statementMapBySubject;

    private final Stack<Object> stack = new Stack<Object>();

    private final List<StatementUnmarshallerFactory> statementUnmarshallerFactories = new ArrayList<StatementUnmarshallerFactory>();

    /*******************************************************************************************************************
     *
     * {@inheritDoc}
     *
     ******************************************************************************************************************/
    @Nonnull
    public <T extends As> T unmarshal (final @Nonnull Graph graph, final @Nonnull Type type)
      throws NotFoundException
      {
        this.graph = graph;

        statementUnmarshallerFactories.addAll(Lookup.getDefault().lookupAll(StatementUnmarshallerFactory.class));
        statementMapBySubject = groupedBySubject(graph);

        for (final Statement statement : findStatements(Vocabulary.RDF_TYPE, type))
          {
            return (T)find(statement.getSubject());
          }

        throw new NotFoundException("Can't find an object of type " + type);
      }

    /*******************************************************************************************************************
     *
     * {@inheritDoc}
     *
     ******************************************************************************************************************/
    @Nonnull
    public <T extends As> List<T> unmarshalAll (final @Nonnull Graph graph, final @Nonnull Type type)
      {
        this.graph = graph;

        statementUnmarshallerFactories.addAll(Lookup.getDefault().lookupAll(StatementUnmarshallerFactory.class));
        statementMapBySubject = groupedBySubject(graph);
        
        final List<T> result = new ArrayList<T>();

        for (final Statement statement : findStatements(Vocabulary.RDF_TYPE, type))
          {
            result.add((T)find(statement.getSubject()));
          }

        return result;
      }

    /*******************************************************************************************************************
     *
     * {@inheritDoc}
     *
     ******************************************************************************************************************/
    public void registerStatementUnmarshallerFactory (final @Nonnull StatementUnmarshallerFactory factory)
      {
        statementUnmarshallerFactories.add(factory);
      }

    /*******************************************************************************************************************
     *
     *
     ******************************************************************************************************************/
    @Nonnull
    private As internalUnmarshal (final @Nonnull Id id)
      {
        final List<Statement> subjectStatements = statementMapBySubject.get(id);

        if (subjectStatements == null)
          {
            throw new IllegalArgumentException("Wrong id: " + id + " - " + statementMapBySubject.keySet());
          }

        final StatementUnmarshaller entityFactory = findEntityFactory(subjectStatements);
        final As result = entityFactory.unmarshal(subjectStatements, this);
        entities.put(id, result);

        return result;
      }

    /*******************************************************************************************************************
     *
     *
     ******************************************************************************************************************/
    @Nonnull
    public Object find (final @Nonnull Object object)
      {
        Object result = null;
        Id id = null;

        if (object instanceof Id)
          {
            id = (Id)object;
          }
        if (object instanceof URI)
          {
            id = Converter.valueToId((URI)object);
          }
        else
          {
            throw new IllegalArgumentException("Only accept an Id or an URI: " + object);
          }

        result = entities.get(id);

        if (result == null)
          {
            result = internalUnmarshal(id);
          }

        return result;
      }

    /*******************************************************************************************************************
     *
     *
     ******************************************************************************************************************/
    @Nonnull
    private StatementUnmarshaller findEntityFactory (final @Nonnull List<Statement> statements)
      {
        for (final StatementUnmarshallerFactory factory : statementUnmarshallerFactories)
          {
            try
              {
                return factory.findStatementUnmarshallerFor(statements);
              }
            catch (NotFoundException e)
              {
                // ok, try next
              }
          }

        throw new RuntimeException("Can't find a StatementUnmarshallerFactory for " + statements);
      }

    /*******************************************************************************************************************
     *
     *
     ******************************************************************************************************************/
    @Nonnull
    private static SortedMap<Id, List<Statement>> groupedBySubject (final @Nonnull Collection<Statement> statements)
      {
        final SortedMap<Id, List<Statement>> result = new TreeMap<Id, List<Statement>>();

        for (final Statement statement : statements)
          {
            final Id subject = Converter.valueToId(statement.getSubject());
            List<Statement> subStatements = result.get(subject);

            if (subStatements == null)
              {
                subStatements = new ArrayList<Statement>();
                result.put(subject, subStatements);
              }

            subStatements.add(statement);
          }

        return result;
      }

    /*******************************************************************************************************************
     *
     *
     ******************************************************************************************************************/
    public void push (final @Nonnull Object object)
      {
        stack.push(object);
      }

    /*******************************************************************************************************************
     *
     *
     ******************************************************************************************************************/
    public void pop()
      {
        stack.pop();
      }

    /*******************************************************************************************************************
     *
     *
     ******************************************************************************************************************/
    @Nonnull
    public Object peek()
      {
        return stack.peek();
      }

    /*******************************************************************************************************************
     *
     *
     ******************************************************************************************************************/
    @Nonnull
    public Collection<Statement> findStatements (final @Nonnull Id predicate, final @Nonnull StringValue object)
      {
        final List<Statement> result = new ArrayList<Statement>();

        for (final Statement statement : graph)
          {
            if (statement.getPredicate().stringValue().equals(predicate.stringValue()))
              {
                if (matches(object, statement.getObject()))
                  {
                    result.add(statement);
                  }
              }
          }

        return result;
      }

    /*******************************************************************************************************************
     *
     *
     ******************************************************************************************************************/
    public boolean matches (final @Nonnull StringValue value, final @Nonnull Object object)
      {
        if (value == null)
          {
            return true;
          }

        if (object instanceof URI)
          {
            if (((URI)object).stringValue().equals(value.stringValue()))
              {
                return true;
              }
          }
        else if (object instanceof Literal)
          {
            if (((Literal)object).stringValue().equals(value.stringValue()))
              {
                return true;
              }
          }

        return false;
      }
  }
